# -*- coding: iso-8859-1 -*-
"""
    MoinMoin - ImprovedTableParser library

    @copyright: 2012 by Paul Boddie <paul@boddie.org.uk>
    @license: GNU GPL (v2 or later), see COPYING.txt for details.
"""

from MoinMoin import wikiutil
from shlex import shlex
from StringIO import StringIO
import re

# Regular expressions.

syntax = {
    # For section markers.
    "markers"   : (r"^\s*(?P<n>\\+)(?P<b>{|})(?P=n)(?P=b)(?P=n)(?P=b)", re.MULTILINE),
    "marker"    : (r"(\\+)",                                            0),

    # At start of line:
    "sections"  : (r"(^\s*{{{.*?^\s*}}})",  re.MULTILINE | re.DOTALL),  # {{{ ... }}}
    "rows"      : (r"^==",                  re.MULTILINE),              # ==

    # Within text:
    "columns"   : (r"\|\|[ \t]*",           0),                         # || ws-excl-nl

    # At start of column text:
    "column"    : (r"^\s*<(.*?)>\s*(.*)",   re.DOTALL),                 # ws < attributes > ws
    }

patterns = {}
for name, (value, flags) in syntax.items():
    patterns[name] = re.compile(value, re.UNICODE | flags)

# Functions.

def parse(s):

    "Parse 's', returning a table definition."

    s = replaceMarkers(s)

    table_attrs = {}
    rows = []

    # The following will be redefined upon the construction of the first row.

    row_attrs = {}
    columns = []

    # Process exposed text and sections.

    exposed = True

    # Initially, start a new row.

    row_continued = False

    for region in patterns["sections"].split(s):

        # Only look for table features in exposed text.

        if exposed:

            # Extract each row from the definition.

            for row_text in patterns["rows"].split(region):

                # Only create a new row when a boundary has been found.

                if not row_continued:
                    if columns:
                        extractAttributes(columns[0][0], row_attrs, table_attrs)

                    row_attrs = {}
                    columns = []
                    rows.append((row_attrs, columns))
                    column_continued = False

                # Extract each column from the row.

                for text in patterns["columns"].split(row_text):

                    # Only create a new column when a boundary has been found.

                    if not column_continued:

                        # Extract the attribute and text sections.

                        match = patterns["column"].search(text)
                        if match:
                            attribute_text, text = match.groups()
                            columns.append([parseAttributes(attribute_text, True), text])
                        else:
                            columns.append([{}, text])

                    else:
                        columns[-1][1] += text

                    # Permit columns immediately following this one.

                    column_continued = False

                # Permit a continuation of the current column.

                column_continued = True

                # Permit rows immediately following this one.

                row_continued = False

            # Permit a continuation if the current row.

            row_continued = True

        # Write any section into the current column.

        else:
            columns[-1][1] += region

        exposed = not exposed

    if columns:
        extractAttributes(columns[0][0], row_attrs, table_attrs)

    return table_attrs, rows

def extractAttributes(attrs, row_attrs, table_attrs):

    """
    Extract row- and table-level attributes from 'attrs', storing them in
    'row_attrs' and 'table_attrs' respectively.
    """

    for name, value in attrs.items():
        if name.startswith("row") and name != "rowspan":
            row_attrs[name] = value
            del attrs[name]
        elif name.startswith("table"):
            table_attrs[name] = value
            del attrs[name]

def replaceMarkers(s):

    "Convert the section notation in 's'."

    l = []
    last = 0

    # Get each marker and convert it.

    for match in patterns["markers"].finditer(s):
        start, stop = match.span()
        l.append(s[last:start])

        # Convert the marker.

        marker = []
        brace = True
        for text in patterns["marker"].split(match.group()):
            if brace:
                marker.append(text)
            else:
                marker.append(text[:-1])
            brace = not brace

        l.append("".join(marker))
        last = stop
    else:
        l.append(s[last:])

    return "".join(l)

def parseAttributes(s, escape=True):

    """
    Parse the table attributes string 's', returning a mapping of names to
    values. If 'escape' is set to a true value, the attributes will be suitable
    for use with the formatter API.
    """

    attrs = {}
    f = StringIO(s)
    name = None
    need_value = False

    for token in shlex(f):

        # Capture the name if needed.

        if name is None:
            name = escape and wikiutil.escape(token) or token

        # Detect either an equals sign or another name.

        elif not need_value:
            if token == "=":
                need_value = True
            else:
                attrs[name.lower()] = escape and "true" or True
                name = wikiutil.escape(token)

        # Otherwise, capture a value.

        else:
            # Quoting of attributes done similarly to parseAttributes.

            if escape and token:
                if token[0] in ("'", '"'):
                    token = wikiutil.escape(token)
                else:
                    token = '"%s"' % wikiutil.escape(token, 1)
    
            attrs[name.lower()] = token
            name = None
            need_value = False

    return attrs

# Formatting of embedded content.
# NOTE: Borrowed from EventAggregator.

def getParserClass(request, format):

    """
    Return a parser class using the 'request' for the given 'format', returning
    a plain text parser if no parser can be found for the specified 'format'.
    """

    try:
        return wikiutil.searchAndImportPlugin(request.cfg, "parser", format or "plain")
    except wikiutil.PluginMissingError:
        return wikiutil.searchAndImportPlugin(request.cfg, "parser", "plain")

def formatText(text, request, fmt):

    "Format the given 'text' using the specified 'request' and formatter 'fmt'."

    parser_cls = getParserClass(request, request.page.pi["format"])
    parser = parser_cls(text, request, line_anchors=False)
    return request.redirectedOutput(parser.format, fmt, inhibit_p=True)

# Common formatting functions.

def formatTable(text, request, fmt):

    "Format the given 'text' using the specified 'request' and formatter 'fmt'."

    attrs, table = parse(text)

    request.write(fmt.table(1, attrs))

    for row_attrs, columns in table:
        request.write(fmt.table_row(1, row_attrs))

        for column_attrs, column_text in columns:
            request.write(fmt.table_cell(1, column_attrs))
            request.write(formatText(column_text, request, fmt))
            request.write(fmt.table_cell(0))

        request.write(fmt.table_row(0))

    request.write(fmt.table(0))

# vim: tabstop=4 expandtab shiftwidth=4
